import os
import sys
import inspect
import datetime
import warnings

from copy import copy
from contextlib import contextmanager

import pkg_resources

import openhtf as htf

from openhtf.util import conf
from openhtf.plugs import user_input, BasePlug

import webbrowser

from ..storage import SITE_DATA_DIR
from ..callbacks.file_provider import TemporaryFileProvider
from ..callbacks.local_storage import LocalStorageOutput


try:
    import tornado, sockjs
    tornado_version = pkg_resources.get_distribution("tornado").version
    major, *rest = tornado_version.split('.')
    if int(major) > 4:
        raise ValueError('Tornado version must be <= 4.x.x')
except (ImportError, ValueError):
    warnings.warn(
        'Tornado 4.x.x not available. GUI server will not work. '
        'If you wish to install it, you may install spintop-openhtf '
        'with the [server] dependency, such as '
        'pip install spintop-openhtf[server]. Do note that this requires '
        'tornado<5.0.'
    )
else:
    from ..callbacks import station_server

from .. import (
    Test,
    # load_component_file,
    # CoverageAnalysis
)

HISTORY_BASE_PATH = os.path.join(SITE_DATA_DIR, 'openhtf-history')

DEFAULT = object()

class TestPlanError(Exception): pass

class TestSequence(object):
    def __init__(self, name):
        self._setup_phases = []
        self._test_phases = []
        self._teardown_phases = []
        self.name = name
    
    def setup(self, name, **options):
        """Decorator factory for a setup function.
        
        ```python
        my_sequence = TestSequence('Parent')
        
        @my_sequence.setup('my-setup-name')
        def setup_fn(test):
            (...)
        
        ```
        """
        return self._decorate_phase(name, self._setup_phases, options)
    
    def testcase(self, name, **options):
        """Decorator factory for a normal phase. 
        """
        return self._decorate_phase(name, self._test_phases, options)
    
    def teardown(self, name, **options):
        """Decorator factory for a teardown phase. 
        """
        return self._decorate_phase(name, self._teardown_phases, options)
    
    def plug(self, *args, **kwargs):
        """Helper method: shortcut to htf.plugs.plug(...)"""
        return htf.plugs.plug(*args, **kwargs)
    
    def measures(self, *args, **kwargs):
        """Helper method: shortcut to htf.measures(...)"""
        return htf.plugs.plug(*args, **kwargs)
    
    def sub_sequence(self, name):
        """Create new empty TestSequence and append it to this sequence.
        
        The following two snippets are equivalent:
        
        ```python
        my_sequence = TestSequence('Parent')
        sub_sequence = my_sequence.sub_sequence('Child')
        ```
        
        ```python
        my_sequence = TestSequence('Parent')
        sub_sequence = TestSequence('Child')
        my_sequence.append(sub_sequence)
        ```
        """
        group = TestSequence(name)
        self.append(group)
        return group
    
    def append(self, phase):
        self._test_phases.append(phase)
        
    def _decorate_phase(self, name, array, options=None):
        if not options:
            options = {}
        options['name'] = name
        def _note_fn(fn):
            phase = self._add_phase(fn, array, options)
            return phase
        return _note_fn
    
    def _add_phase(self, fn, array, options):
        phase = ensure_htf_phase(fn)
        phase.options.update(**options)
        array.append(phase)
        return phase
        
    @property
    def phase_group(self):
        # Recursively get phase groups of sub phases if available, else the phase itself.
        _test_phases = [getattr(phase, 'phase_group', phase) for phase in self._test_phases]
        
        return htf.PhaseGroup(
            setup=self._setup_phases,
            main=_test_phases,
            teardown=self._teardown_phases,
            name=self.name
        )
    
class TestPlan(TestSequence):
    DEFAULT_CONF = dict(
        station_server_port='4444', 
        capture_docstring=True
    )
    def __init__(self, name='testplan', store_result=True):
        super(TestPlan, self).__init__(name=name)
        
        self._execute_test = None
        self._top_level_component = None
        self.coverage = None
        self.file_provider = TemporaryFileProvider()
        self.callbacks = []
        
        # Array but must contain only one phase.
        # Array is for compatibility with self._decorate_phase function of parent class.
        self._trigger_phases = []
        self._no_trigger = False
        
        if store_result:
            self.add_callbacks(LocalStorageOutput(_local_storage_filename_pattern, indent=4))
            
        self.failure_exceptions = (user_input.SecondaryOptionOccured,)

    @property
    def execute_test(self):
        """Returns a function that takes no arguments and that executes the test described by this test plan."""
        if not self._execute_test:
            self.freeze_test()
        return self._execute_test
    
    @property
    def history_path(self):
        path = os.path.join(HISTORY_BASE_PATH, self.name)
        if not os.path.exists(path):
            os.makedirs(path)
        return path
    
    @property
    def is_runnable(self):
        phases = self._test_phases + self._trigger_phases
        return bool(phases)
    
    @property
    def trigger_phase(self):
        return self._trigger_phases[0] if self._trigger_phases else None
        
    @property
    def is_test_frozen(self):
        return bool(self._execute_test)
        
    def image_url(self, url):
        return self.file_provider.create_url(url)
    
    def trigger(self, name, **options):
        """Decorator factory for the trigger phase. 
        
        Similar to `testcase`, except that this function will be used as the test trigger.
        
        The test trigger is a special test phase that is executed before test officialy start.
        Once this phase is complete, the test will start. Usually used to configure the test with
        the DUT id for example."""
        if self.trigger_phase:
            raise TestPlanError('There can only be one @trigger function.')
        
        return self._decorate_phase(name, self._trigger_phases, options)

    def no_trigger(self):
        self._no_trigger = True
    
    def add_callbacks(self, *callbacks):
        if self.is_test_frozen:
            raise RuntimeError('Cannot add callbacks to the test plan after the test was frozen.')
        self.callbacks += callbacks
    
    def assert_runnable(self):
        if not self.is_runnable:
            # No phases ! Abort now.
            raise RuntimeError('Test is empty, aborting.')
    
    def execute(self):
        """ Execute the configured test using the test_start function as a trigger.
        """
        return self.execute_test()
    
    def run_once(self, launch_browser=True):
        return self.run(launch_browser=launch_browser, once=True)
    
    def run(self, launch_browser=True, once=False):
        self._load_default_conf()
        with self._station_server_context(launch_browser):
            while True:
                try:
                    self.execute()
                except KeyboardInterrupt:
                    break
                finally:
                    if once:
                        break
    
    def freeze_test(self):
        self._execute_test = self._create_execute_test()
        
    @contextmanager
    def _station_server_context(self, launch_browser=True):
        with station_server.StationServer(self.file_provider) as server:
            self.add_callbacks(server.publish_final_state)
            self.assert_runnable() # Check before launching browser
            self.freeze_test()
            
            if launch_browser and conf['station_server_port']:
                webbrowser.open('http://localhost:%s' % conf['station_server_port'])
            
            yield
    
    def _load_default_conf(self):
        conf.load_from_dict(self.DEFAULT_CONF, _override=False)

    def _create_execute_test(self):
        self.assert_runnable()
        
        test = Test(self.phase_group, test_name=self.name, _code_info_after_file=__file__)
        test.configure(failure_exceptions=self.failure_exceptions)
        test.add_output_callbacks(*self.callbacks)
        
        
        trigger_phase = self._build_trigger_phase()
        
        def execute_test():
            return test.execute(test_start=trigger_phase)

        return execute_test
        
    def _build_trigger_phase(self):
        if self.trigger_phase:
            return self.trigger_phase
        elif self._no_trigger:
            return None
        else:
            return create_default_trigger()
    
    def create_plug(self):
        class _SelfReferingPlug(BasePlug):
            def __new__(cls):
                return self
        
        return _SelfReferingPlug
    
    def spintop_plug(self, fn):
        return htf.plugs.plug(spintop=self.create_plug())(fn)
    

def _local_storage_filename_pattern(**test_record):
    folder = '{metadata[test_name]}'.format(**test_record)
    start_time_datetime = datetime.datetime.utcfromtimestamp(test_record['start_time_millis']/1000.0)
    start_time = start_time_datetime.strftime(r"%Y_%m_%d_%H%M%S_%f")
    subfolder = '{dut_id}_{start_time}_{outcome}'.format(start_time=start_time, **test_record)
    return os.path.join(HISTORY_BASE_PATH, folder, subfolder)
    
def create_default_trigger(message='Enter a DUT ID in order to start the test.',
        validator=lambda sn: sn, **state):
    
    @htf.PhaseOptions(timeout_s=None, requires_state=True)
    @htf.plugs.plug(prompts=user_input.UserInput)
    def trigger_phase(state, prompts):
        """Test start trigger that prompts the user for a DUT ID."""
        dut_id = prompts.prompt(message, text_input=True)
        state.test_record.dut_id = validator(dut_id)
        
    return trigger_phase
    
def ensure_htf_phase(fn):
    if not hasattr(fn, 'options'):
        # Not a htf phase, decorate it so it becomes one.
        fn = htf.PhaseOptions()(fn) 
    return fn



import pytest
import mock

from openhtf.core import test_record

from spintop_openhtf.util.gdrive import FileNotFound, load_credentials_file, GoogleDrive, Worksheet, GoogleFolder

from spintop_openhtf.callbacks.gdrive import TestRecordToSheetOutput

def create_test_record(config={}):
     return test_record.TestRecord(
        dut_id='test-device', station_id='unit-testing-station', code_info=test_record.CodeInfo(name=None, docstring=None, sourcecode=None),
        start_time_millis=0,
        metadata={'config': config}
     )
     
@pytest.mark.skip
def test_folder_created():
    test_record = create_test_record()

    with mock.patch('spintop_openhtf.util.gdrive.GoogleDrive') as GoogleDrive, \
            mock.patch('spintop_openhtf.util.gdrive.Worksheet') as Worksheet, \
            mock.patch('spintop_openhtf.util.gdrive.GoogleFolder') as GoogleFolder:

        worksheet = Worksheet.return_value
        folder = GoogleFolder()
        output = TestRecordToSheetOutput(folder)
        output(test_record)
        folder.get_or_create_spreadsheet.assert_called()

        Worksheet.assert_called()

        print(folder.get_or_create_spreadsheet.return_value)
        Worksheet.assert_called_with(folder.get_or_create_spreadsheet.return_value)


    


import logging

from spintop_openhtf import conf, plugs

DEFAULT_FORMAT = '%(asctime)-15s: %(message)s'

class UnboundPlug(object):
    def __init__(self):
        super().__init__()
        if not hasattr(self, 'logger'):
            # Check conditionnaly because openhtf BasePlug will raise an
            # exception if we define self.logger.
            self.logger = logging.getLogger(str(id(self)))

    def tearDown(self):
        """Tear down the plug instance."""
        self.close()
        self.logger.close_log()

    def open(self, *args, **kwargs):
        raise NotImplementedError()

    def close(self):
        raise NotImplementedError()

    def log_to_stream(self, stream=None, **kwargs):
        handler = logging.StreamHandler(stream)
        self._add_handler(handler, **kwargs)

    def log_to_filename(self, filename, **kwargs):
        handler = logging.FileHandler(filename)
        self._add_handler(handler, **kwargs)

    def _add_handler(self, handler, level=logging.DEBUG, format=DEFAULT_FORMAT):
        handler.setLevel(level)
        if format:
            handler.setFormatter(logging.Formatter(format))
        self.logger.addHandler(handler)

    def close_log(self):
        for handler in self.logger.handlers:
            handler.close()

    @classmethod
    def as_plug(cls, name, **kwargs_values):
        """ Create a bound plug that will retrieve values from conf or the passed values here.

        Take SSHInterface for example. 

        ```python
        from spintop_openhtf.plugs import from_conf, SSHInterface

        MySSHInterface = SSHInterface.as_plug(
            'MySSHInterface', # The name of this plug as it will appear in logs
            addr=from_conf( # from_conf will retrieve the conf value named like this.
                'my_ssh_addr',
                description="The addr of my device."
            ), 
            username='x', # Always the same
            password='y'  # Always the same
        )

        ```
        """
        return plug_factory(cls, name, **kwargs_values)


def plug_factory(plug_cls, name, **kwargs_values):

    for key, value in kwargs_values.items():
        if isinstance(value, from_conf):
            value.declare()
            
    def resolve(value):
        if isinstance(value, from_conf):
            return value.resolve()
        else:
            return value


    class _AnonymousPlug(plugs.BasePlug, plug_cls):
        def __init__(self):
            kwargs = {key: resolve(value) for key, value in kwargs_values.items()}
            super().__init__(**kwargs)

    _AnonymousPlug.__name__ = name
    return _AnonymousPlug

class from_conf(object):
    def __init__(self, conf_name, description=None):
        self.conf_name = conf_name
        self.description = description
    
    def resolve(self):
        return conf[self.conf_name]

    def declare(self):
        conf.declare(self.conf_name, description=self.description)


import pytest
import os
import mock

from collections import namedtuple

from oauth2client.service_account import ServiceAccountCredentials

import openhtf as htf
from spintop_openhtf.callbacks.gdrive import create_google_drive_outputs, load_credentials_file_with_scope
from spintop_openhtf.util.gdrive import GoogleDrive



HERE = os.path.abspath(os.path.dirname(__file__))

CREDS_FILENAME = os.path.expanduser(os.path.join('~', 'tack-infra-af44c5fee082.json'))# os.getenv('GS_CREDS_FILENAME')
DUMMY_ACCESS_TOKEN = '<ACCESS_TOKEN>'


class DummyCredentials(object):
    def __init__(self, dummy_token):
        self.access_token = dummy_token

    def authorize(self, *args, **kwargs):
        pass

# import vcr
# vcr = vcr.VCR(
#     filter_headers=['authorization'],
#     cassette_library_dir=os.path.join(HERE, 'cassettes')
# )

def sanitize_token(interaction, current_cassette):
    headers = interaction.data['request']['headers']
    token = headers.get('Authorization')

    if token is None:
        return

    interaction.data['request']['headers']['Authorization'] = [
        'Bearer %s' % DUMMY_ACCESS_TOKEN
    ]


def get_credentials():
    if CREDS_FILENAME:
        credentials = load_credentials_file_with_scope(CREDS_FILENAME)
    else:
        credentials = DummyCredentials(DUMMY_ACCESS_TOKEN)
    
    return credentials

@pytest.fixture()
def gwrap():
    credentials = get_credentials()
    gwrap = GoogleDrive(credentials)
    return gwrap
